#include <winsock2.h>
#include <WS2tcpip.h> // pour inet_pton()
#include <iostream>
#include <string>
using namespace std;

#pragma comment(lib,"Ws2_32.lib")

using socket_t = SOCKET;
bool est_invalide(socket_t sck) {
   return sck == INVALID_SOCKET;
}
using socket_operation_t = int;
bool est_erreur(socket_operation_t op) {
   return op == SOCKET_ERROR;
}

template <class T>
   void print(ostream &os, T && arg) {
      os << arg;
   }
template <class T, class ... Args>
   void print(ostream &os, T && arg, Args && ... args) {
      print(os, std::forward<T>(arg));
      print(os, std::forward<Args>(args)...);
   }

template <class ... Args>
   auto erreur(Args && ... args) {
      const auto code = WSAGetLastError();
      print(cerr, std::forward<Args>(args)...);
      cerr << "; erreur " << code << endl;
      return code;
   }

void charger_sockets() noexcept {
   const long VERSION_DEMANDEE = MAKEWORD(2, 2);
   WSAData configWinsock;
   if (WSAStartup(VERSION_DEMANDEE, &configWinsock) != 0)
      exit(erreur("Demarrage"));
}

void decharger_sockets() noexcept {
   WSACleanup();
}

socket_t creer_socket() {
   auto sck = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
   if (est_invalide(sck))
      exit(erreur("Creation"));
   return sck;
}

void reserver_port(socket_t sck, unsigned short port) {
   sockaddr_in nom = { 0 };
   nom.sin_family = AF_INET;
   nom.sin_port = htons(port);
   auto resultat = bind(
      sck, reinterpret_cast<const sockaddr*>(&nom), sizeof(sockaddr_in)
   );
   if (est_erreur(resultat))
      exit(erreur("Port ", port, ", obtention"));
   // Crer la file dattente pour demandes de connexion
   int taille_file = 16; // arbitraire
   if (est_erreur(listen(sck, taille_file)))
      exit(erreur("Creer file d'attente de taille ", taille_file));
}

socket_t accepter_client(socket_t srv) {
   sockaddr_in homologue = { 0 };
   int taille = sizeof(sockaddr_in);
   auto sck = accept(
      srv, reinterpret_cast<sockaddr *>(&homologue), &taille
   );
   if (est_invalide(sck))
      exit(erreur("Accepter connexion"));
   return sck;
}

string recevoir(socket_t sck, char fin) {
   const int CAPACITE = 64; // arbitraire
   char tampon[CAPACITE];
   string message_recu;
   bool reception_completee = false;
   do {
      auto n = recv(sck, tampon, CAPACITE, 0);
      if (est_erreur(n))
         exit(erreur("Rception"));
      message_recu.append(tampon + 0, tampon + n);
      if (tampon[n-1] == fin)
         reception_completee = true;
   } while (!reception_completee);
   if (!message_recu.empty()) message_recu.pop_back(); // liminer le dlimiteur
   return message_recu;
}

void envoyer(socket_t sck, string msg, char fin) {
   msg += fin;
   while (!msg.empty()) {
      auto plus_recent= send(
         sck, &msg[0], static_cast<int>(msg.size()), 0
      );
      if (est_erreur(plus_recent))
         exit(erreur("Envoi"));
      msg = msg.substr(plus_recent);
   }
}

void fermer_socket(socket_t sck) noexcept {
   shutdown(sck, SD_BOTH);
   closesocket(sck);
}


int main() {
   const unsigned short PORT_A_RESERVER= 4321;
   const string MESSAGE = "Recu 5 sur 5";
   const char DELIMITEUR_FIN = '#';
   charger_sockets();
   auto serveur = creer_socket();
   reserver_port(serveur, PORT_A_RESERVER);
   auto pourJaser = accepter_client(serveur);
   auto recu = recevoir(pourJaser, DELIMITEUR_FIN);
   cout << "Reception reussie de " << recu.size() + 1
        << " bytes, delimiteur inclus. Message : \"" << recu << "\"\n";
   envoyer(pourJaser, MESSAGE, DELIMITEUR_FIN);
   cout << "Envoi reussi de " << MESSAGE.size() + 1
        << " bytes, delimiteur inclus. Message : \"" << MESSAGE << '\"' << endl;
   fermer_socket(pourJaser);
   fermer_socket(serveur);
   decharger_sockets();
}
